@@ -33,6 +33,10 @@ use constant CHUNKSIZE    => 64 * 1024;
 use constant DEBUG        => $ENV{MONOCEROS_DEBUG} || 0;
 
 my $null_io = do { open my $io, "<", \""; $io };
+my $have_accept4 = eval {
+    require Linux::Socket::Accept4;
+    Linux::Socket::Accept4::SOCK_CLOEXEC()|Linux::Socket::Accept4::SOCK_NONBLOCK();
+};
 
 sub new {
     my $class = shift;
@@ -631,15 +635,22 @@ sub accept_or_recv {
     my $self = shift;
     my @for_read = @_;
     my $conn;
+    use open 'IO' => ':unix';
     for my $sock ( @for_read ) {
         if ( fileno $sock == fileno $self->{listen_sock} ) {
-            my $peer = accept(my $fh, $self->{listen_sock});
+            my ($fh,$peer);
+            if ( $have_accept4 ) {
+                $peer = Linux::Socket::Accept4::accept4($fh,$self->{listen_sock}, $have_accept4);
+            }
+            else {
+                $peer = accept($fh,$self->{listen_sock});
+                fh_nonblocking($fh,1) if $peer;
+            }
             if ( !$peer && ($! != EINTR && $! != EAGAIN && $! != EWOULDBLOCK && $! != ESPIPE) ) {
                 warn sprintf 'failed to accept: %s (%d)', $!, $!;
                 next;
             }
             next unless $peer;
-            fh_nonblocking($fh,1);
             setsockopt($fh, IPPROTO_TCP, TCP_NODELAY, 1)
                 or die "setsockopt(TCP_NODELAY) failed:$!";
             my ($peerport,$peerhost) = unpack_sockaddr_in $peer;
